@@ -1,5 +1,5 @@
 import re
-from typing import Union
+from typing import Union, Dict
 
 import js2py
 import js2py.internals.simplex
@@ -11,35 +11,36 @@ from extreme_parser.amazon.product.model import Product
 from extreme_parser.util.parse import parse_number
 
 
-def product_overview_feature_div_table(selector: parsel.Selector, field: str) -> str:
-    table = selector.xpath("//div[@id='productOverview_feature_div']//table").get()
-    if table is None:
+def table(selector: parsel.Selector, exp: str, *fields) -> str:
+    tbl = selector.xpath(exp).get()
+    if tbl is None:
         return ""
-    table = pandas.read_html(table)
-    if len(table) <= 0:
+    tbl = pandas.read_html(tbl)
+    if len(tbl) <= 0:
         return ""
-    table = table[0].set_index(0)
-    if field not in table[1]:
+    tbl = tbl[0].set_index(0)
+
+    for f in fields:
+        if f in tbl[1]:
+            return tbl[1][f]
+    else:
         return ""
-    s = table[1][field]
-    return s
 
 
-def product_details_table(selector: parsel.Selector, field: str) -> str:
-    table = selector.xpath(
-        "//table[@id='productDetails_detailBullets_sections1']|"
-        "//table[@id='productDetails_techSpec_section_1']"
-    ).get()
-    if table is None:
-        return ""
-    table = pandas.read_html(table)
-    if len(table) <= 0:
-        return ""
-    table = table[0].set_index(0)
-    if field not in table[1]:
-        return ""
-    s = table[1][field]
-    return s
+def tabular_buy_box_container(selector: parsel.Selector, *fields) -> str:
+    return table(selector, "//table[@id='tabular-buybox-container']", *fields)
+
+
+def product_overview_feature_div(selector: parsel.Selector, *fields) -> str:
+    return table(selector, "//div[@id='productOverview_feature_div']//table", *fields)
+
+
+def product_details_detail_bullets(selector: parsel.Selector, *fields) -> str:
+    return table(selector, "//table[@id='productDetails_detailBullets_sections1']", *fields)
+
+
+def product_details_tech_spec(selector: parsel.Selector, *fields) -> str:
+    return table(selector, "//table[@id='productDetails_techSpec_section_1']", *fields)
 
 
 def parse(html: str, p: Product):
@@ -53,6 +54,8 @@ def parse(html: str, p: Product):
     parse_stock(p, selector=sel)
     parse_ship_day(p, selector=sel)
     parse_ship(p, selector=sel)
+    parse_sold(p, selector=sel)
+    parse_offers(p, selector=sel)
     parse_delivery(p, selector=sel)
     parse_material(p, selector=sel)
     parse_description(p, selector=sel)
@@ -60,6 +63,9 @@ def parse(html: str, p: Product):
     parse_size(p, selector=sel)
     parse_color(p, selector=sel)
     parse_variation(p, selector=sel)
+    parse_review_amount(p, selector=sel)
+    parse_review_star(p, selector=sel)
+    parse_rank(p, selector=sel)
 
 
 def parse_product_asin(p: Product, selector: parsel.Selector = None):
@@ -72,7 +78,9 @@ def parse_parent_asin(p: Product, selector: parsel.Selector = None):
 
 def parse_weight(p: Product, selector: parsel.Selector = None):
     def s1() -> str:
-        return product_details_table(selector, "Item Weight")
+        s = product_details_detail_bullets(selector, "Item Weight")
+        s = s or product_details_tech_spec(selector, "Item Weight")
+        return s
 
     def s2() -> str:
         ul = selector.xpath("//div[@id='detailBullets_feature_div']/ul")
@@ -83,26 +91,22 @@ def parse_weight(p: Product, selector: parsel.Selector = None):
         return s
 
     weight_str = s1() or s2()
-    if weight_str == "":
-        p.weight = None
-        return
     weight: Union[None, int, float] = parse_number(weight_str)
     if weight is None:
         p.weight = None
         return
 
-    if weight_str.endswith("ounces") or weight_str.endswith("Ounces"):
-        p.weight = weight * 0.0625
-    elif weight_str.endswith("pounds") or weight_str.endswith("Pounds"):
-        p.weight = float(weight)
+    if weight_str.lower().endswith("ounces"):
+        p.weight = round(weight * 0.0625, 2)
+    elif weight_str.lower().endswith("pounds"):
+        p.weight = round(float(weight), 2)
     else:
         p.weight = None
-        return
 
 
 def parse_brand(p: Product, selector: parsel.Selector = None):
     def s1() -> str:
-        return product_overview_feature_div_table(selector, "Brand")
+        return product_overview_feature_div(selector, "Brand")
 
     def s2() -> str:
         s = selector.xpath("//a[@id='bylineInfo']/text()").re_first("Brand: (.*)")
@@ -118,7 +122,7 @@ def parse_brand(p: Product, selector: parsel.Selector = None):
 
 
 def parse_price(p: Product, selector: parsel.Selector = None):
-    def table() -> pandas.DataFrame:
+    def price_table() -> pandas.DataFrame:
         t = pandas.DataFrame(index=("origin", "discount"), columns=('title', 'price'))
         for i, tr in enumerate(selector.xpath("//div[@id='price']/table/tr")[:2]):
             t.iat[i, 0] = tr.xpath("normalize-space(./td[1]/text())").get()
@@ -133,7 +137,7 @@ def parse_price(p: Product, selector: parsel.Selector = None):
         return s or ""
 
     def s2() -> str:
-        t = table()
+        t = price_table()
         if numpy.all(t["price"] == ""):
             return ""
         if numpy.any(t["price"] == ""):
@@ -201,17 +205,7 @@ def parse_ship_day(p: Product, selector: parsel.Selector = None):
 
 def parse_ship(p: Product, selector: parsel.Selector = None):
     def s1() -> str:
-        table = selector.xpath("//table[@id='tabular-buybox-container']").get()
-        if table is None:
-            return ""
-        table = pandas.read_html(table)
-        if len(table) <= 0:
-            return ""
-        table = table[0].set_index(0)
-        if "Ships from" not in table[1]:
-            return ""
-        s = table[1]["Ships from"]
-        return s
+        return tabular_buy_box_container(selector, "Ships from")
 
     def s2() -> str:
         ship = selector.xpath("normalize-space(//div[@id='sfsb_accordion_head']/div/div/span[2]/text())").get()
@@ -233,6 +227,23 @@ def parse_ship(p: Product, selector: parsel.Selector = None):
         p.ship = None
 
 
+def parse_sold(p: Product, selector: parsel.Selector = None):
+    p.sold = selector.xpath(
+        "normalize-space(//span[contains(text(), 'Sold by:')]/following-sibling::*[1]/text())"
+    ).get()
+
+
+def parse_offers(p: Product, selector: parsel.Selector = None):
+    offers = selector.xpath(
+        "//div[@id='olpLinkWidget_feature_div']//div[@class='olp-text-box']/span[1]/text()"
+    ).get()
+    if offers is None:
+        p.offers = None
+        return
+
+    p.offers = parse_number(offers)
+
+
 def parse_delivery(p: Product, selector: parsel.Selector = None):
     delivery = selector.xpath("string(//div[@id='contextualIngressPtLabel_deliveryShortLine'])").get()
     delivery = delivery or selector.xpath("normalize-space(//span[@id='glow-ingress-line2']/text())").get()
@@ -254,7 +265,7 @@ def parse_delivery(p: Product, selector: parsel.Selector = None):
 
 
 def parse_material(p: Product, selector: parsel.Selector = None):
-    material = product_overview_feature_div_table(selector, "Material")
+    material = product_overview_feature_div(selector, "Material")
     if material == "":
         p.material = None
     else:
@@ -279,8 +290,8 @@ def parse_title(p: Product, selector: parsel.Selector = None):
 
 
 def parse_size(p: Product, selector: parsel.Selector = None):
-    dimensions = product_details_table(selector, "Product Dimensions")
-    dimensions = dimensions or product_details_table(selector, "Package Dimensions")
+    dimensions = product_details_detail_bullets(selector, "Product Dimensions", "Package Dimensions")
+    dimensions = dimensions or product_details_tech_spec(selector, "Product Dimensions", "Package Dimensions")
     if not dimensions.endswith("inches"):
         p.size_length = None
         p.size_width = None
@@ -339,18 +350,52 @@ def parse_variation(p: Product, selector: parsel.Selector = None):
     def change_key(d: dict):
         return {labels[k]: v for k, v in d.items() if k in labels}
 
-    def change_variations(vars_d: dict) -> list:
-        vars_l = list()
-        for _, var in vars_d.items():
+    def change_variations(vars_ori: dict) -> Dict[str, dict]:
+        vars_change = dict()
+        for asin, var in vars_ori.items():
             var_change = dict()
             for k, v in var.items():
                 if k in labels and k in values:
                     var_change[labels[k]] = values[k][int(v)]
                 else:
                     var_change[k] = v
-            vars_l.append(var_change)
-        return vars_l
+            vars_change[asin] = var_change
+        return vars_change
 
     p.variation_values = change_key(values)
     p.selected_variations = change_key(selected)
     p.asin_variation_values = change_variations(variations)
+
+
+def parse_review_amount(p: Product, selector: parsel.Selector = None):
+    amount = selector.xpath("//span[@id='acrCustomerReviewText']/text()").get()
+    if amount is None:
+        p.review_amount = None
+    else:
+        p.review_amount = parse_number(amount)
+
+
+def parse_review_star(p: Product, selector: parsel.Selector = None):
+    star = selector.xpath("//span[@id='acrPopover']/@title").get()
+    if star is None:
+        p.review_star = None
+    else:
+        p.review_star = parse_number(star)
+
+
+def parse_rank(p: Product, selector: parsel.Selector = None):
+    rank = product_details_detail_bullets(selector, "Best Sellers Rank")
+    if rank == "":
+        p.rank = None
+        return
+
+    rank = re.sub(r"\(.*?\)", "", rank)
+    rank = re.findall(r"([\d,]+) in ([^\d]+)", rank)
+    if len(rank) == 0:
+        p.rank = None
+        return
+
+    rank = {r[1].strip(): int(r[0].replace(",", ""))
+            for r in rank
+            if len(r) == 2 and r[0].replace(",", "").isdigit()}
+    p.rank = rank
